#!/usr/bin/env python

import numpy as np

from evaluator.dataset import ZEWDPCBaseDataset, ZEWDPCProtectedDataset

import torch 
import torch.nn as nn
import torch.optim as optim
import torch.nn.functional as F
from torch.utils.data import Dataset, DataLoader, ConcatDataset, random_split
import pandas as pd
import numpy as np
import time
import copy
import math
from torchvision import models, transforms
import random
from sklearn.metrics import precision_score, accuracy_score, multilabel_confusion_matrix


class ZEWDPCBaseRun:
    """
    Template Submission Class for the ZEW Data Purchasing Challenge 2022.

    The submission template follows the following hooks :
        - pre_training_phase
        - purchase_phase
        - prediction_phase
        - save_checkpoint
        - load_checkpoint

        Please refer to the inline documentation for further details.
        You are allowed to add any other member functions, however you
        are not allowed to change the names of these hooks, otherwise your
        submissions will not be evaluated by the automated evaluators.
    """

    def train_model(self,model, dataloader : DataLoader,criterion, optimizer : optim.Optimizer,num_epochs = 30,maxtime = 3600,validation = None, scheduler = None, early_stopping_rounds = 10):

        start = time.time()
        acc_history = []
        loss_history = []

        best_model_wts = copy.deepcopy(model.state_dict())
        best_loss = math.inf

        if validation == None:
            validation = dataloader
        stop = early_stopping_rounds

        model = model.to(self.device)

        for epoch in range(num_epochs):

            torch.set_grad_enabled(True)

            if time.time() - start > maxtime:
                model.load_state_dict(best_model_wts)
                return model, acc_history, loss_history

            # Train phase
            model.train()
            torch.set_grad_enabled(True)
            running_loss = 0
            running_correct = 0

            for data in dataloader:
                images = data['image'].to(self.device)
                labels = torch.transpose(torch.stack(data['label']),0,1).to(self.device)
                optimizer.zero_grad()
                predicted = model(images)
                loss = criterion(predicted,labels.float())
                loss.backward()
                optimizer.step()
                running_loss += loss.item()
                running_correct += torch.sum(((predicted > 0) == labels).all(dim=1)).item()

            loss_history.append(running_loss)
            acc_history.append(running_correct/len(dataloader.dataset))

            model.eval()

            if scheduler != None:
                scheduler.step()

            torch.set_grad_enabled(False)
            if validation != None: 
                temperature, validation_loss = self.calculate_temperature(model,validation)

                print(epoch, " Validation Loss: ", validation_loss)

                stop = stop - 1
                if validation_loss < best_loss:
                    stop = early_stopping_rounds
                    best_loss = validation_loss
                    #best_running_incorrect = running_incorrect
                    best_model_wts = copy.deepcopy(model.state_dict())
                if stop <= 0:
                    model.load_state_dict(best_model_wts)
                    return model, acc_history, loss_history
            else:
                if best_loss > loss_history[-1]:
                    best_loss = loss_history[-1]
                    best_model_wts = copy.deepcopy(model.state_dict())

        model.load_state_dict(best_model_wts)
        return model, acc_history, loss_history

    def __init__(self):
        self.evaluation_state = {}

        self.device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")

        self.LOCAL = True
        self.BATCH_SIZE = 16
        self.TRAIN_TEST_SPLIT_PERCENTAGE = 0.9

        self.model = models.efficientnet_b1(pretrained=True)

        self.model.classifier[1] = nn.Linear(1280,4)

        self.general_train_transform = transforms.Compose([
            transforms.ToTensor(),
            transforms.RandomHorizontalFlip(p=0.5),
            transforms.RandomVerticalFlip(p=0.5),
            #transforms.ColorJitter(brightness=.2,hue=.2),
            self.CustomRotation(angles=[-90, 0, 90, 180]),
            transforms.Resize(224),
            transforms.Normalize([0.485, 0.456, 0.406], [0.229, 0.224, 0.225]),
        ])

        self.general_validate_transform = transforms.Compose([
                transforms.ToTensor(),
                transforms.Resize(224),
                self.CustomRotation(angles=[-90, 0, 90, 180]),
                transforms.Normalize([0.485, 0.456, 0.406], [0.229, 0.224, 0.225]),
        ])

    def pre_training_phase(
        self, training_dataset: ZEWDPCBaseDataset, register_progress=lambda x: False
    ):
        print("\n================> Pre-Training Phase\n")

        # Data Setup
        train_raw, test_raw = random_split(training_dataset,[int(self.TRAIN_TEST_SPLIT_PERCENTAGE*len(training_dataset)),len(training_dataset)-int(self.TRAIN_TEST_SPLIT_PERCENTAGE*len(training_dataset))],generator=torch.Generator().manual_seed(42))
        self.train = self.TransformDatasetWrapper(train_raw,self.general_train_transform)
        test_a = self.TransformDatasetWrapper(test_raw,self.general_validate_transform)
        test_b = self.TransformDatasetWrapper(test_a,transforms.functional.hflip)
        test_c = self.TransformDatasetWrapper(test_a,transforms.functional.vflip)
        test_d = self.TransformDatasetWrapper(test_a,transforms.Compose([
            transforms.functional.hflip,
            transforms.functional.vflip
        ]))
        self.test = ConcatDataset((test_a,test_b,test_c,test_d))

        # Class distribution
        self.class_distribution = torch.from_numpy(np.sum(training_dataset._get_all_labels(),axis=0) / len(training_dataset)).to(self.device)

        # Pretrain
        training_dataloader = DataLoader(self.train,batch_size=self.BATCH_SIZE)
        validation_dataloader = DataLoader(self.test,batch_size=self.BATCH_SIZE)

        optimizer = optim.Adam(self.model.parameters())
        criterion = nn.BCEWithLogitsLoss()
        scheduler = torch.optim.lr_scheduler.MultiplicativeLR(optimizer, lambda epoch: 0.92 ** epoch)

        self.model, acc_history, loss_history = self.train_model(self.model,training_dataloader,criterion,optimizer,validation=validation_dataloader,num_epochs=10,early_stopping_rounds=3, maxtime = 2400, scheduler = scheduler)

        print("Execution Complete of Training Phase.")


    def purchase_phase(
        self,
        unlabelled_dataset: ZEWDPCProtectedDataset,
        training_dataset: ZEWDPCBaseDataset,
        budget=1000,
        register_progress=lambda x: False,
    ):

        validation_dataloader = DataLoader(self.test,batch_size=self.BATCH_SIZE)
        temperature,_ = self.calculate_temperature(self.model,validation_dataloader)
        model = self.model
        device = self.device

        model.eval()
        predicted_all = None
        correct_all = None
        set_temp = False
        for data in validation_dataloader:
            images = data['image']
            labels = torch.transpose(torch.stack(data['label']),0,1)

            labels = labels.to(device).cpu().detach().numpy()
            images = images.to(device)

            predicted = model(images).cpu().detach().numpy()

            if not set_temp:
                set_temp = True
                predicted_all = predicted
                correct_all = labels
            else:
                predicted_all = np.concatenate((predicted_all,predicted))
                correct_all = np.concatenate((correct_all,labels))

        confusion_matrix = multilabel_confusion_matrix(correct_all,predicted_all > 0)

        accuracy_vector = [
                            (confusion_matrix[0][0][0] + confusion_matrix[0][1][1]) / len(validation_dataloader.dataset),
                            (confusion_matrix[1][0][0] + confusion_matrix[1][1][1]) / len(validation_dataloader.dataset),
                            (confusion_matrix[2][0][0] + confusion_matrix[2][1][1]) / len(validation_dataloader.dataset),
                            (confusion_matrix[3][0][0] + confusion_matrix[3][1][1]) / len(validation_dataloader.dataset)
                            ] 
        print(accuracy_vector)

        register_progress(0.1)

        #######################################################################################

        unlabelled_predicted = []
        unlabelled_dataset.set_transform(self.general_validate_transform)
        unlabelled_dataloader = DataLoader(unlabelled_dataset,batch_size=self.BATCH_SIZE)

        torch.no_grad()
        model.eval()

        threshold = 0.8

        for data in unlabelled_dataloader:
            images = data['image'].to(device)
            
            predicted = model(images)
            confidence_values = torch.prod(torch.sigmoid(self.scale_temperature(torch.abs(predicted),temperature)),1)
            predicted = predicted.cpu().detach().numpy()
            for i in range(len(predicted)):
                unlabelled_predicted.append({
                    'idx': int(data['idx'][i]),
                    'prediction': predicted[i],
                    'label': (predicted[i] > 0).astype(int).tolist(),
                    'image': None
                })
        
        register_progress(0.2)

        def purchase_score(sample):
            # Can probably be improved a lot
            prediction = self.sigmoid(sample['prediction'])
            res = 0
            for i, x in enumerate(prediction):
                res += (1 - accuracy_vector[i]) * prediction[i]
            return res
    
        unlabelled_predicted.sort(key = purchase_score, reverse = True)

        for i in range(min(budget,len(unlabelled_predicted))):
            unlabelled_predicted[i]['label'] = unlabelled_dataset.purchase_label(unlabelled_predicted[i]['idx'])

        unlabelled_dataset.set_transform(None)
        purchase_dataset = self.MapDataset(unlabelled_dataset,unlabelled_predicted[:min(budget,len(unlabelled_predicted))])
        remaining_dataset = self.MapDataset(unlabelled_dataset,unlabelled_predicted[min(budget,len(unlabelled_predicted)):])

        register_progress(0.3)
        
        remaining_dataset.set_transform(self.general_validate_transform)
        remaining_dataloader = DataLoader(remaining_dataset,batch_size=self.BATCH_SIZE)

        selected_pseudo = []

        torch.no_grad()
        model.eval()

        threshold = 0.7
        for data in remaining_dataloader:
            images = data['image'].to(device)
            
            predicted = model(images)
            confidence_values = torch.prod(torch.sigmoid(self.scale_temperature(torch.abs(predicted),temperature)),1)
            
            chosen = confidence_values > threshold
            pseudo_labels = (predicted[chosen] > 0)

            pseudo_labels = pseudo_labels.cpu().detach().numpy()
            for i in range(len(pseudo_labels)):
                selected_pseudo.append({
                    'idx': int(data['idx'][chosen][i]),
                    'label': (pseudo_labels[i]).astype(int).tolist(),
                    'image': None
                })

        unlabelled_dataset.set_transform(None)
        pseudo_dataset = self.MapDataset(unlabelled_dataset,selected_pseudo)

        pseudo_dataset.set_transform(self.general_train_transform)
        purchase_dataset.set_transform(self.general_train_transform)
        full_dataset = ConcatDataset([pseudo_dataset,self.train,purchase_dataset])
        full_dataloader = DataLoader(full_dataset,batch_size = self.BATCH_SIZE,shuffle=True)

        current_distribution = torch.zeros(4).to(device)
        for data in full_dataloader:
            current_distribution = current_distribution + torch.sum(torch.transpose(torch.stack(data['label']),0,1).to(device),dim=0)

        weights = self.class_distribution / (current_distribution / len(full_dataset)+1e-5)

        model = models.efficientnet_b1(pretrained=True)
        model.classifier[1] = nn.Linear(1280,4)

        optimizer = optim.Adam(model.parameters(),weight_decay=1e-5)
        scheduler = torch.optim.lr_scheduler.MultiplicativeLR(optimizer, lambda epoch: 0.92 ** epoch)
        criterion = nn.BCEWithLogitsLoss(pos_weight = weights)

        self.model, acc_history, loss_history = self.train_model(model,full_dataloader,criterion,optimizer,validation=validation_dataloader,num_epochs=40,early_stopping_rounds=10, scheduler = scheduler)

    

    def prediction_phase(
        self,
        test_dataset: ZEWDPCBaseDataset,
        register_progress=lambda x: False,
    ):
        """
        # Prediction Phase
        -------------------------
        In this phase of the competition, you have access to the test dataset, and you
        are supposed to make predictions using your trained models.

        Returns:
            np.ndarray of shape (n, 4)
                where n is the number of samples in the test set
                and 4 refers to the 4 labels to be predicted for each sample
                for the multi-label classification problem.

        PARTICIPANT_TODO: Add your code here
        """
        print(
            "\n================> Prediction Phase : - on {} images\n".format(
                len(test_dataset)
            )
        )
        
        predicted_all, correct_all = self.evaluate(self.model,test_dataset,has_label = False)

        return predicted_all > 0


    # ======================================== UTIL ======================================== #

    # Adapted from https://github.com/gpleiss/temperature_scaling/blob/master/temperature_scaling.py

    class CustomRotation:
        """Rotate by one of the given angles."""

        def __init__(self, angles):
            self.angles = angles

        def __call__(self, x):
            angle = random.choice(self.angles)
            return transforms.functional.rotate(x, angle)

    def scale_temperature(self,logits,temperature):
        return torch.div(logits,temperature)

    def calculate_temperature(self,model,dataloader):
        # Returns 4 temperature values
        criterion = nn.BCEWithLogitsLoss()
        logits = []
        labels = []

        model.eval()
        with torch.no_grad():
            for sample in dataloader:
                labels.append(torch.transpose(torch.stack(sample['label']),0,1))
                logits.append(model(sample['image'].to(self.device)))
            logits = torch.cat(logits).to(self.device)
            labels = torch.cat(labels).to(self.device).float()

        initial_loss = criterion(logits,labels).item()
        temperature = torch.ones(4,requires_grad=True,device=self.device)
        optimizer = optim.LBFGS([temperature], lr=0.05, max_iter=200)

        def eval():
            optimizer.zero_grad()
            loss = criterion(self.scale_temperature(logits,temperature),labels)
            loss.backward()
            return loss
        optimizer.step(eval)

        return temperature.detach(), initial_loss

    def evaluate(self,model,dataset,has_label = True):
        model.eval()
        predicted_all = None
        correct_all = None
        set_temp = False
        dataset.set_transform(self.general_validate_transform)
        dataloader_a = DataLoader(dataset,shuffle=False)
        dataloader_b = DataLoader(self.TransformDatasetWrapper(dataset,transforms.functional.hflip),shuffle=False)
        dataloader_c = DataLoader(self.TransformDatasetWrapper(dataset,transforms.functional.vflip),shuffle=False)
        dataloader_d = DataLoader(self.TransformDatasetWrapper(dataset,lambda x : transforms.functional.rotate(x,90)),shuffle=False)
        dataloader_e = DataLoader(self.TransformDatasetWrapper(dataset,lambda x : transforms.functional.rotate(x,270)),shuffle=False)

        dataloader_iterator_b = iter(dataloader_b)
        dataloader_iterator_c = iter(dataloader_c)
        dataloader_iterator_d = iter(dataloader_d)
        dataloader_iterator_e = iter(dataloader_e)
        # Apparently zip has a memory leakage issue
        for i, s_a in enumerate(dataloader_a):
            s_b = next(dataloader_iterator_b)
            s_c = next(dataloader_iterator_c)
            s_d = next(dataloader_iterator_d)
            s_e = next(dataloader_iterator_e)

            im_a = s_a['image'].to(self.device)
            im_b = s_b['image'].to(self.device)
            im_c = s_c['image'].to(self.device)
            im_d = s_d['image'].to(self.device)
            im_e = s_e['image'].to(self.device)

            if has_label:
                labels = torch.transpose(torch.stack(s_a['label']),0,1)
                labels = labels.to(self.device).cpu().detach().numpy()

            predicted = np.mean(
                np.array(
                    [
                        ((model(im_a) > 0) * 2 - 1).cpu().detach().numpy(),
                        ((model(im_b) > 0) * 2 - 1).cpu().detach().numpy(),
                        ((model(im_c) > 0) * 2 - 1).cpu().detach().numpy(),
                        ((model(im_d) > 0) * 2 - 1).cpu().detach().numpy(),
                        ((model(im_e) > 0) * 2 - 1).cpu().detach().numpy()
                    ]
                ),
                axis = 0
            )

            if not set_temp:
                set_temp = True
                predicted_all = predicted
                if has_label:
                    correct_all = labels
            else:
                predicted_all = np.concatenate((predicted_all,predicted))
                if has_label:
                    correct_all = np.concatenate((correct_all,labels))
        return predicted_all,correct_all

    def save_checkpoint(self, checkpoint_path):
        
        state_dict = copy.deepcopy(self.model.state_dict())
        save_dict = {
            'state_dict':state_dict,
            'train':self.train,
            'test':self.test,
            'class_distribution': self.class_distribution
        }

        torch.save(save_dict, checkpoint_path)

    def load_checkpoint(self, checkpoint_path):

        save_dict = torch.load(checkpoint_path)

        self.model.load_state_dict(save_dict['state_dict'])
        self.model = self.model.to(self.device)
        self.test = save_dict['test']
        self.train = save_dict['train']
        self.class_distribution = save_dict['class_distribution']

        self.device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")

    def sigmoid (self,x):
        return 1/(1 + np.exp(-x))
    class MapDataset (Dataset):
        def __init__(self,database,arr,transform = None):
            self.arr = arr
            self.transform = transform
            self.database = database

        def __getitem__(self,i):
            sample = self.arr[i]
            im = self.database[sample['idx']]['image']
            if self.transform:
                im = self.transform(im)
            return  {
            'idx':sample['idx'],
            'image':im,
            'label':sample['label']
            }

        def __len__(self):
            return len(self.arr)

        def set_transform(self,transform):
            self.transform = transform 
    class TransformDatasetWrapper (Dataset):
        def __init__(self,database,transform):
            self.transform = transform
            self.database = database

        def __getitem__(self,i):
            sample = self.database[i]
            sample['image'] = self.transform(sample['image'])
            return sample

        def __len__(self):
            return len(self.database)

        def set_transform(self,transform):
            self.transform = transform

if __name__ == "__main__":
    ####################################################################################
    ## You need to implement `ZEWDPCBaseRun` class in this file for this challenge.
    ## Code for running all the phases locally is written in `main.py` for illustration
    ## purposes.
    ##
    ## Checkout the inline documentation of `ZEWDPCBaseRun` for more details.
    ####################################################################################
    import local_evaluation
